class Entity {
  String _entityId;
  String _resolverAddress;
  String _networkId;
  List<String> _entryPoints;

  // Metadata
  final List<String> languages;
  final Map<String, String> name; // language dependent
  final Map<String, String> description; // language dependent
  final String votingContract;
  final List<String> activeProcessIds;
  final List<String> endedProcessIds;
  final Map<String, String> newsFeed; // language dependent
  final String avatar;
  // final String imageHeader;
  List<EntityAction> actions = [];

  List gatewayBootNodes = []; // TODO: unused by now
  Map<String, dynamic> gatewayUpdate = {}; // TODO: unused
  List relays = []; // TODO: unused by now
  List bootEntities = []; // TODO: unused by now
  List fallbackBootNodeEntities = []; // TODO: unused by now
  List trustedEntities = []; // TODO: unused by now
  List censusServiceManagedEntities = []; // TODO: unused by now

  // From the resolver
  // final String metadataOrigin;

  Entity({
    this.languages,
    this.name,
    this.description,
    // this.metadataOrigin,
    this.votingContract,
    this.activeProcessIds,
    this.endedProcessIds,
    this.newsFeed,
    this.avatar,
    // this.imageHeader,
    this.actions,
    this.gatewayBootNodes,
    this.gatewayUpdate,
    this.relays,
    this.bootEntities,
    this.fallbackBootNodeEntities,
    this.trustedEntities,
    this.censusServiceManagedEntities,
  });

  /// Parses a JSON oject into an Entity instance
  Entity.fromJson(Map<String, dynamic> json)
      : languages = (json['languages'] ?? []).cast<String>().toList(),
        name = Map<String, String>.from(json['name'] ?? {}),
        description = Map<String, String>.from(json['description'] ?? {}),
        votingContract = json['votingContract'],
        newsFeed = Map<String, String>.from(json['newsFeed'] ?? {}),
        activeProcessIds = ((json['process-ids'] ?? {})['active'] ?? [])
            .cast<String>()
            .toList(),
        endedProcessIds = ((json['process-ids'] ?? {})['ended'] ?? [])
            .cast<String>()
            .toList(),
        avatar = json['avatar'],
        // imageHeader = json['imageHeader'],
        actions = (json['actions'] ?? [])
            .map((item) => EntityAction.fromJson(item))
            .cast<EntityAction>()
            .toList(),
        gatewayBootNodes = json['gatewayBootNodes'] ?? [],
        gatewayUpdate = json['gatewayUpdate'] ?? {},
        relays = json['relays'] ?? [],
        bootEntities = json['bootEntities'] ?? [],
        fallbackBootNodeEntities = json['fallbackBootNodeEntities'] ?? [],
        trustedEntities = json['trustedEntities'] ?? [],
        censusServiceManagedEntities =
            json['censusServiceManagedEntities'] ?? [];

  /// Returns a JSON object that can be serialized
  Map<String, dynamic> toJson() {
    return {
      'languages': languages,
      'name': name,
      'description': description,
      // 'meta': metadataOrigin,
      'votingContract': votingContract,
      'newsFeed': newsFeed,
      'process-ids': {'active': activeProcessIds, 'ended': endedProcessIds},
      'avatar': avatar,
      // 'imageHeader': imageHeader,
      'actions': actions.map((EntityAction item) => item.toJson()).toList(),
      'gatewayBootNodes': gatewayBootNodes,
      'gatewayUpdate': gatewayUpdate,
      'relays': relays,
      'bootEntities': bootEntities,
      'fallbackBootNodeEntities': fallbackBootNodeEntities,
      'trustedEntities': trustedEntities,
      'censusServiceManagedEntities': censusServiceManagedEntities
    };
  }

  /// Returns a JSON object that can be serialized, including the
  /// external fields entityId, resolverAddress and networkId
  Map<String, dynamic> toJsonAll() {
    final json = this.toJson();

    json["external"] = {
      "entityId": _entityId,
      "resolverAddress": _resolverAddress,
      "networkId": _networkId,
      "entryPoints": _entryPoints,
    };
    return json;
  }

  // The fields and methods below do not belong to the JSON metadata specification

  /// Set utility fields that do not belong to the JSON metadata specification
  setExternalFields(String entityId, String resolverAddress,
      {String networkId, List<String> entryPoints}) {
    _entityId = entityId;
    _resolverAddress = resolverAddress;
    if (networkId != null) _networkId = networkId;
    if (entryPoints != null) _entryPoints = entryPoints;
  }

  String get entityId => _entityId;
  String get resolverAddress => _resolverAddress;
  String get networkId => _networkId;
  List<String> get entryPoints => _entryPoints;
}

// HELPER CLASSES

/// The base class for entity actions
abstract class EntityAction {
  String type; // browser, image
  Map<String, String> name;
  String visible;

  EntityAction({this.type, this.name, this.visible});

  static EntityAction fromJson(Map<String, dynamic> json) {
    switch (json["type"]) {
      case "browser":
        return EntityActionBrowser.fromJson(json);
      case "image":
        return EntityActionImage.fromJson(json);
      default:
        throw "Invalid action type";
    }
  }

  Map<String, dynamic> toJson();
}

/// Tells the mobile app to navigate to a certain location
class EntityActionBrowser extends EntityAction {
  String url;

  EntityActionBrowser.fromJson(Map<String, dynamic> json)
      : url = json["url"],
        super(
            type: json["type"],
            name: Map<String, String>.from(json['name'] ?? {}),
            visible: json["visible"]);

  @override
  Map<String, dynamic> toJson() {
    return {
      // generic
      "type": this.type,
      "name": this.name,
      "visible": this.visible,
      // custom
      "url": this.url
    };
  }
}

/// App-driven image upload
class EntityActionImage extends EntityAction {
  /// Endpoint accepting POST requests with JSON payloads like:
  /// {
  ///   "field-name-1": "base64-image-payload",
  ///   "another-field": "base64-image-payload",
  ///   ...
  /// }
  ///
  /// The URL will receive the following query string parameters:
  /// - signature = sign(hash(jsonBody), privateKey)
  /// - publicKey
  String url;

  /// Requested image types to provide
  List<ImageUploadSource> imageSources;

  EntityActionImage.fromJson(Map<String, dynamic> json)
      : url = json["url"],
        imageSources = (json["imageSources"] ?? [])
            .map((item) => ImageUploadSource.fromJson(item))
            .cast<ImageUploadSource>()
            .toList(),
        super(
            type: json["type"],
            name: Map<String, String>.from(json['name'] ?? {}),
            visible: json["visible"]);

  @override
  Map<String, dynamic> toJson() {
    return {
      // generic
      "type": this.type,
      "name": this.name,
      "visible": this.visible,
      // custom
      "url": this.url,
      "imageSources": this.imageSources.map((item) => item.toJson()).toList()
    };
  }
}

/// Specific type of image to upload
class ImageUploadSource {
  /// front-camera | back-camera | gallery
  String type;

  /// Arbitrary name to identify the data when the JSON is posted
  String name;

  /// portrait | landscape
  /// Optional when type == "gallery"
  String orientation;

  /// face | id-card-front | id-card-back
  String overlay;
  Map<String, String> caption;

  ImageUploadSource({this.type});

  ImageUploadSource.fromJson(Map<String, dynamic> json)
      : type = json["type"],
        name = json["name"],
        orientation = json["orientation"],
        overlay = json["overlay"],
        caption = Map<String, String>.from(json['caption'] ?? {});

  Map<String, dynamic> toJson() {
    return {
      "type": type,
      "name": name,
      "orientation": orientation,
      "overlay": overlay,
    };
  }
}
